(* Schalter-Komponenten
   - einfacher Kippschalter (an - aus): TTumblerSwitch
   - Momentanschalter                 : TMomSwitch
   - Zweiseitiger Momentanschalter    : TBiMomSwitch
   - Schiebeschalter                  : TSlideSwitch

    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Mrz. 2007
   Vers. 2 - Apr. 2015: TSlideSwitch (DelphiXE2 adoption of koro_smoothshift)
   Vers. 2.1 - Sep. 2015: TToggleSwitch renamed to TTumblerSwitch to prevent
             from conflict with new TToggleSwitch component of Delphi 10            
   *)

unit Switches;

interface

uses
  System.SysUtils, System.Classes, Vcl.Controls, Winapi.Windows,
  Winapi.Messages, Vcl.Graphics;

// Definiere CompPalPage (siehe Register)
{$Include UserComps.pas }

type
  TSwitchState = (ssOn,ssOff);
  TSwitchShape = (ssRect,ssRound);
  TSwitchPosition = (spOff,spNeutral,spOn);

  TCustomSwitch = class(TCustomControl)
  private
    procedure DoEnter; override;
    procedure DoExit; override;
  protected
    FOuterColor,
    FInnerColor,
    FSwOffColor,
    FSwOnColor  : TColor;
    procedure SetOuterColor (Value : TColor);
    procedure SetInnerColor (Value : TColor);
    procedure SetOffColor (Value : TColor);
    procedure SetOnColor (Value : TColor);
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Anchors;
    property Enabled;
    property OuterColor : TColor read FOuterColor write SetOuterColor default clSilver;
    property InnerColor : TColor read FInnerColor write SetInnerColor default clGray;
    property SwitchOffColor : TColor read FSwOffColor write SetOffColor default clWhite;
    property SwitchOnColor : TColor read FSwOnColor write SetOnColor default clWhite;
    property Visible;
  end;

  TCustomTumblerSwitch = class(TCustomSwitch)
  protected
    FPosition : TSwitchPosition;
    FSwNeutralColor : TColor;
    FShape : TSwitchShape;
    procedure SetNeutralColor (Value : TColor);
    procedure SetShape (Value : TSwitchShape);
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property NeutralColor : TColor read FSwNeutralColor write SetNeutralColor default clWhite;
    property Shape : TSwitchShape read FShape write SetShape default ssRound;
    property TabOrder;
    property TabStop;
  end;

  TTumblerSwitch = class(TCustomTumblerSwitch)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
    FState : TSwitchState;
    procedure SetState (Value : TSwitchState);
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property State : TSwitchState read FState write SetState default ssOff;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

  TMomSwitch = class(TCustomTumblerSwitch)
  private
    FOnClick : TNotifyEvent;
    FRestPos : TSwitchPosition;
    procedure SetRestPos (Value : TSwitchPosition);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property RestPosition : TSwitchPosition read FRestPos write SetRestPos default spOff;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
  end;

  TBiMomSwitch = class(TCustomTumblerSwitch)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { Published-Deklarationen }
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

  TGlyphHeight = (gh32,gh48,gh64,gh96);

  TCustomSwitch3D = class(TCustomControl)
  private
    FGlyphHeight : TGlyphHeight;
    FBitMap     : TBitMap;
    FPosition   : TSwitchPosition;
    procedure SetGlyphHeight (Value : TGlyphHeight);
    procedure SetPosition (Value : TSwitchPosition);
    procedure AdjustSize (var W, H: Integer); reintroduce;
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
  protected
    procedure Paint; override;
    property GlyphHeight : TGlyphHeight read FGlyphHeight write SetGlyphHeight default gh32;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    property Anchors;
    property Enabled;
    property Visible;
    end;

  TSwitch3D = class(TCustomSwitch3D)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
    FState : TSwitchState;
    procedure SetState (Value : TSwitchState);
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property GlyphHeight;
    property State : TSwitchState read FState write SetState default ssOff;
    property TabOrder;
    property TabStop;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
    end;

  TMomSwitch3D = class(TCustomSwitch3D)
  private
    FOnClick : TNotifyEvent;
    FRestPos : TSwitchPosition;
    procedure SetRestPos (Value : TSwitchPosition);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property RestPosition : TSwitchPosition read FRestPos write SetRestPos default spOff;
    property TabOrder;
    property TabStop;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
  end;

  TBiMomSwitch3D = class(TCustomSwitch3D)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { Published-Deklarationen }
    property TabOrder;
    property TabStop;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

  TSlideSwitch = class(TCustomSwitch)
  private
    FMouseHover : boolean;
    FliderRect : TRect;
    FHoverColor : TColor;
    FState : TSwitchState;
    FSliderRect : TRect ;
    FCaptionOn,FCaptionOff : string;
    FOnOnClick : TNotifyEvent;
    FOnOffClick : TNotifyEvent;
    procedure SetHoverColor (Value : TColor);
    procedure SetCaptionOn(Value : string);
    procedure SetCaptionOff(Value : string);
    procedure SetMouseHover(value : boolean);
    procedure SetState (Value : TSwitchState);
    procedure UpdateSliderRect;
    procedure CMFocusChanged(var message : TCMFocusChanged); message CM_FOCUSCHANGED;
    procedure CMEnabledChanged(var message : TMessage); message CM_ENABLEDCHANGED;
    procedure WMWindowPosChanged(var message : TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure CMMouseLeave(var message : TMessage); message CM_MOUSELEAVE;
    procedure CMDialogChar(var message : TCMDialogChar); message CM_DIALOGCHAR;
  protected
    procedure Paint; override;
    procedure MouseMove(Shift : TShiftState; x,y : integer); override;
    procedure MouseDown(Button : TMouseButton; Shift : TShiftState; x,y : integer); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner : TComponent); override;
    procedure ToggleState;
  published
    property Font;
    property CaptionOn : string read FCaptionOn write SetCaptionOn;
    property CaptionOff : string read FCaptionOff write SetCaptionOff;
    property HoverColor : TColor read FHoverColor write SetHoverColor default clSkyBlue;
    property State : TSwitchState read FState write SetState default ssOff;
    property OnKeyDown;
    property OnKeyUp;
    property OnOffClick: TNotifyEvent read FOnOffClick write FOnOffClick;
    property OnOnClick: TNotifyEvent read FOnOnClick write FOnOnClick;
  end;

procedure Register;

implementation

{$R Switch.res}

uses Vcl.Themes, Winapi.UxTheme;

constructor TCustomSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width:=16;
  Height:=16;
  FOuterColor:=clSilver;
  FInnerColor:=clGray;
  FSwOffColor:=clWhite;
  FSwOnColor:=clWhite;
  end;

procedure TCustomSwitch.SetOuterColor (Value : TColor);
begin
  FOuterColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetInnerColor (Value : TColor);
begin
  FInnerColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetOffColor (Value : TColor);
begin
  FSwOffColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetOnColor (Value : TColor);
begin
  FSwOnColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.DoEnter;
begin
  inherited;
  Invalidate;
  end;

procedure TCustomSwitch.DoExit;
begin
  inherited;
  Invalidate;
  end;

//====================================================================
constructor TCustomTumblerSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width:=16;
  Height:=20;
  FPosition:=spOff;
  FSwNeutralColor:=clWhite;
  end;

procedure TCustomTumblerSwitch.SetNeutralColor (Value : TColor);
begin
  FSwNeutralColor:=Value;
  Invalidate;
  end;

procedure TCustomTumblerSwitch.SetShape (Value : TSwitchShape);
begin
  FShape:=Value;
  Invalidate;
  end;

procedure TCustomTumblerSwitch.Paint;
var
  n,d,h,m : integer;
begin
  with Canvas do begin
    if Focused then Brush.Color:=FInnerColor
    else Brush.Color:=FOuterColor;
    with Pen do begin
      Color:=clBlack; Width:=0;
      end;
    m:=Height div 10;
    h:=Height-Height div 5;
    Ellipse (0,m,h,m+h);
    if Focused then Brush.Color:=FOuterColor
    else Brush.Color:=FInnerColor;
    n:=h div 8;
    Ellipse (n,m+n,h-n,m+h-n);
    Pen.Width:=n div 2;
    n:=h div 3;
    Brush.Color:=clBlack;
    Ellipse (n,m+n,h-n,m+h-n);
    m:=h div 2;
    n:=Height div 2; d:=Height div 8;
    case FPosition of
    spOff : begin
             Brush.Color:=FSwOffColor;
             if FShape=ssRect then RoundRect(m-d,n-d,m+d,Height-1,d,d)
             else Ellipse (m-d,n-d,m+d,Height-1);
             end;
    spOn   : begin
             Brush.Color:=FSwOnColor;
             if FShape=ssRect then RoundRect(m-d,n+d,m+d,1,d,d)
             else Ellipse (m-d,n+d,m+d,0);
             end;
      else begin
         Brush.Color:=clWhite;
         Ellipse (m-d,n-d,m+d,n+d);
        end;
      end;
    end;
  end;

//====================================================================
constructor TTumblerSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FState:=ssOff;
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TTumblerSwitch.SetState (Value : TSwitchState);
begin
  if FState<>Value then begin
    FState:=Value;
    if FState=ssOn then FPosition:=spOn else FPosition:=spOff;
    Invalidate;
    end;
  end;

procedure TTumblerSwitch.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    ord('U'),ord('u'),VK_ADD: State:=ssOn;
    ord('D'),ord('d'),VK_SUBTRACT : State:=ssOff;
    end;
  end;

procedure TTumblerSwitch.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if State=ssOff then begin
    State:=ssOn;
    if assigned(FOnUpClick) then FOnUpClick(self);
    end
  else begin
    State:=ssOff;
    if assigned(FOnDownClick) then FOnDownClick(self);
    end;
  end;

//====================================================================
constructor TMomSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPosition:=spOff;
  FOnClick:=nil;
  end;

procedure TMomSwitch.SetRestPos (Value : TSwitchPosition);
begin
  FRestPos:=Value;
  FPosition:=Value;
  Invalidate;
  end;

procedure TMomSwitch.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if FRestPos=spOff then FPosition:=spOn else FPosition:=spOff;
  Invalidate;
  if assigned(FOnClick) then FOnClick(self);
  end;

procedure TMomSwitch.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  FPosition:=FRestPos;
  Invalidate;
  end;

//====================================================================
constructor TBiMomSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPosition:=spNeutral;
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TBiMomSwitch.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if y>Height div 2 then begin
    FPosition:=spOff;
    if assigned(FOnDownClick) then FOnDownClick(self);
    end
  else begin
    FPosition:=spOn;
    if assigned(FOnUpClick) then FOnUpClick(self);
    end;
  Invalidate;
  end;

procedure TBiMomSwitch.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  FPosition:=spNeutral;
  Invalidate;
  end;

//====================================================================
(* 3D Schalter *)
type
  TSwitchSize = record
    GWidth,GHeight : integer;
    end;

const
  GlyphHeights : array [TGlyphHeight] of TSwitchSize =
     ((GWidth:21;GHeight:32),(GWidth:32;GHeight:48),(GWidth:43;GHeight:64),(GWidth:64;GHeight:96));

constructor TCustomSwitch3D.Create (AOwner: TComponent);
begin
  FBitMap:=TBitMap.Create;
  with FBitMap do begin
    Transparent:=True;
    TransparentMode:=tmAuto;
    Height:=32;
    end;
  inherited Create(AOwner);
  GlyphHeight:=gh32;
  FPosition:=spOff;
  end;

destructor TCustomSwitch3D.Destroy;
begin
  FBitMap.Free;
  inherited Destroy;
  end;

procedure TCustomSwitch3D.SetGlyphHeight (Value : TGlyphHeight);
var
  s : string;
begin
  FGlyphHeight:=Value;
  case Value of
  gh48 : if FPosition=spOff then s:='SwitchU48'
         else if FPosition=spOn then s:='SwitchO48'
         else s:='SwitchM48';
  gh64 : if FPosition=spOff then s:='SwitchU64'
         else if FPosition=spOn then s:='SwitchO64'
         else s:='SwitchM64';
  gh96 : if FPosition=spOff then s:='SwitchU96'
         else if FPosition=spOn then s:='SwitchO96'
         else s:='SwitchM96';
    else if FPosition=spOff then s:='SwitchU32'
         else if FPosition=spOn then s:='SwitchO32'
         else s:='SwitchM32';
    end;
  FBitMap.Handle:=LoadBitmap(HInstance,pchar(s));
  with GlyphHeights[Value] do begin
    Width:=GWidth; Height:=GHeight;
    end;
  end;

procedure TCustomSwitch3D.SetPosition (Value : TSwitchPosition);
begin
  FPosition:=Value;
  SetGlyphHeight(FGlyphHeight);
  Invalidate;
  end;

procedure TCustomSwitch3D.AdjustSize(var W,H: Integer);
begin
  if (csLoading in ComponentState) then Exit;
  with FBitMap do begin
    if W<Width then W:=Width;
    if H<Height then H:=Height;
    end;
  end;

procedure TCustomSwitch3D.SetBounds(ALeft,ATop,AWidth,AHeight: Integer);
var
  W, H: Integer;
begin
  W:=AWidth;
  H:=AHeight;
  AdjustSize(W,H);
  inherited SetBounds(ALeft,ATop,W,H);
  end;

procedure TCustomSwitch3D.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;
  { check for minimum size }
  W:=Width; H:=Height;
  AdjustSize(W,H);
  if (W<>Width) or (H<>Height) then inherited SetBounds(Left,Top,W,H);
  Message.Result:=0;
  end;

procedure TCustomSwitch3D.Paint;
begin
  Canvas.Draw(0,0,FBitMap);
  end;

//====================================================================
constructor TSwitch3D.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FState:=ssOff;
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TSwitch3D.SetState (Value : TSwitchState);
begin
  FState:=Value;
  if FState=ssOn then FPosition:=spOn else FPosition:=spOff;
  SetGlyphHeight(FGlyphHeight);
  Invalidate;
  end;

procedure TSwitch3D.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    ord('U'),ord('u'),VK_ADD: State:=ssOn;
    ord('D'),ord('d'),VK_SUBTRACT : State:=ssOff;
    end;
  end;

procedure TSwitch3D.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if State=ssOff then begin
    State:=ssOn;
    if assigned(FOnUpClick) then FOnUpClick(self);
    end
  else begin
    State:=ssOff;
    if assigned(FOnDownClick) then FOnDownClick(self);
    end;
  end;

//====================================================================
constructor TMomSwitch3D.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetPosition(spOff);
  FOnClick:=nil;
  end;

procedure TMomSwitch3D.SetRestPos (Value : TSwitchPosition);
begin
  FRestPos:=Value;
  FPosition:=Value;
  Invalidate;
  end;

procedure TMomSwitch3D.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if FRestPos=spOff then SetPosition(spOn) else SetPosition(spOff);
  Invalidate;
  if assigned(FOnClick) then FOnClick(self);
  end;

procedure TMomSwitch3D.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  FPosition:=FRestPos;
  Invalidate;
  end;

//====================================================================
constructor TBiMomSwitch3D.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetPosition(spNeutral);
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TBiMomSwitch3D.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if y>Height div 2 then begin
    SetPosition(spOff);
    if assigned(FOnDownClick) then FOnDownClick(self);
    end
  else begin
    SetPosition(spOn);
    if assigned(FOnUpClick) then FOnUpClick(self);
    end;
  Invalidate;
  end;

procedure TBiMomSwitch3D.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  SetPosition(spNeutral);
  Invalidate;
  end;

//====================================================================
const
  sOFF = 'OFF';
  sON  = 'ON';

constructor TSlideSwitch.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Width:=50;
  Height:=20;
  FOuterColor:=clGray;
  FInnerColor:=clSilver;
  FHoverColor:=clSkyBlue;
  FSwOffColor:=clRed;
  FSwOnColor:=clGreen;
  ControlStyle:=[csCaptureMouse,csClickEvents,csOpaque];
  fCaptionOff:=sOff;
  fCaptionOn:=sOn;
  Caption:=fCaptionOff;
  with Font do begin
    Name:='Tahoma';
    Style:=[fsBold];
    Color:=clWhite;
    Size:=8;
    end;
  Canvas.Font:=Font;
  UpdateSliderRect;
  end;

procedure TSlideSwitch.SetMouseHover(value : boolean);
begin
  if fMouseHover<>Value then begin
    fMouseHover:=value;
    invalidate;
    end;
  end;

procedure TSlideSwitch.UpdateSliderRect;
begin
  if FState=ssOn then SetRect(fSliderRect,round(3/5*Width),0,Width,Height)
  else SetRect(fSliderRect,0,0,round(2/5*Width),Height)
  end;

procedure TSlideSwitch.CMEnabledChanged(var message : TMessage);
begin
  invalidate;
  inherited;
  end;

procedure TSlideSwitch.CMFocusChanged(var message: TCMFocusChanged);
begin
  invalidate;
  inherited;
  end;

procedure TSlideSwitch.CMMouseLeave(var message : TMessage);
begin
  SetMouseHover(false);
  inherited;
  end;

procedure TSlideSwitch.WMWindowPosChanged(var message : TWMWindowPosChanged);
begin
  inherited;
  UpdateSliderRect;
  Font.Size:=Round(height div 3)+1;
  end;

procedure TSlideSwitch.CMDialogChar(var message : TCMDialogChar);
begin
  with message do
  // ...
    begin
      Click;
      Result:=1;
      inherited;
    end;
end;

procedure TSlideSwitch.SetState (Value : TSwitchState);
begin
  if FState<>Value then begin
    FState:=Value;
    if FState=ssOn then caption:=fCaptionOn  else Caption:=fCaptionOff;
    UpdateSliderRect;
    Invalidate;
    end;
  end;

procedure TSlideSwitch.ToggleState;
begin
  if FState=ssOn then SetState(ssOff) else SetState(ssOn);
  end;

procedure TSlideSwitch.KeyDown(var key : word; shift : TShiftState);
begin
  case Key of
  ord('L'),ord('l'),VK_LEFT: SetState(ssOn);
  ord('R'),ord('r'),VK_RIGHT: SetState(ssOff);
  VK_SPACE : ToggleState;
    end;
  inherited KeyUp(Key,shift);
  end;

procedure TSlideSwitch.SetHoverColor (Value : TColor);
begin
  FHoverColor:=Value;
  Invalidate;
  end;

procedure TSlideSwitch.SetCaptionOn(Value : string);
begin
  fCaptionOn:=Value;
  if FState=ssOn then begin
    Caption:=fCaptionOn;
    invalidate;
    end;
  end;

procedure TSlideSwitch.SetCaptionOff(Value : string);
begin
  fCaptionOff:=Value;
  if FState=ssOff then begin
    Caption:=fCaptionOff;
    invalidate;
    end;
  end;

procedure TSlideSwitch.MouseDown(button : TMouseButton; shift : TShiftState;
  x,y: integer);
begin
  if (shift = [ssLeft]) and PtInRect(fSliderRect,Point(x,y)) then ToggleState;
  inherited MouseDown(button,shift,x,y);
  if (FState=ssOff) then begin
    if assigned(FOnOffClick) then FOnOffClick(self);
    end
  else if assigned(FOnOnClick) then FOnOnClick(self);
  end;

procedure TSlideSwitch.MouseMove(shift: TShiftState; x,y: Integer);
begin
  if GetCaptureControl=nil then SetMouseHover(PtInRect(fSliderRect,Point(x,y)));
  inherited MouseMove(shift,x,y);
  end;

type
  TStyleTextFlagEx = (stfTextColor, stfBorderColor, stfBorderSize,
    stfShadowColor, stfShadowOffset, stfShadowType, stfGlowSize);
  TStyleTextFlagsEx = set of TStyleTextFlagEx;

  TStyleTextOptionsEx = record
    Flags: TStyleTextFlagsEx;
    TextColor: TColor;
    BorderColor: TColor;
    BorderSize: Integer;
    ShadowColor: TColor;
    ShadowOffset: TPoint;
    ShadowType,
    GlowSize: Integer;
  end;

const
  tstNone = TST_NONE;
  tstSingle = TST_SINGLE;
  tstContinuous = TST_CONTINUOUS;

function DrawTextEx(ThemeServices : TCustomStyleServices; DC: HDC; Details: TThemedElementDetails;
  const S: string; var R: TRect; Flags: TTextFormat; Options: TStyleTextOptionsEx): Boolean;
const
  COptions: array[TStyleTextFlagEx] of Cardinal = (DTT_TEXTCOLOR, DTT_BORDERCOLOR,
    DTT_BORDERSIZE, DTT_SHADOWCOLOR, DTT_SHADOWOFFSET, DTT_SHADOWTYPE, DTT_GLOWSIZE);
var
  LFlags: Cardinal;
  LOptions: TDTTOpts;
  LTextOption: TStyleTextFlagEx;
  LColorRef: TColorRef;
begin
  LFlags := TTextFormatFlags(Flags);

  // Vista or later, use DrawThemeTextEx
  if TOSVersion.Check(6) then
  begin
    FillChar(LOptions, SizeOf(LOptions), 0);
    LOptions.dwSize := SizeOf(LOptions);
    for LTextOption := Low(TStyleTextFlagEx) to High(TStyleTextFlagEx) do
      if (LTextOption in Options.Flags) then
        LOptions.dwFlags := LOptions.dwFlags or COptions[LTextOption];
    LOptions.crText := Vcl.Graphics.ColorToRGB(Options.TextColor);
    LOptions.crBorder := Vcl.Graphics.ColorToRGB(Options.BorderColor);
    LOptions.iBorderSize := Options.BorderSize;
    LOptions.crShadow := Vcl.Graphics.ColorToRGB(Options.ShadowColor);
    LOptions.ptShadowOffset := Options.ShadowOffset;
    LOptions.iTextShadowType:=Options.ShadowType;      // JR - 2015-04-24
    LOptions.iGlowSize := Options.GlowSize;
    if (tfComposited in Flags) then
      LOptions.dwFlags := LOptions.dwFlags or DTT_COMPOSITED;
    if (tfCalcRect in Flags) then
      LOptions.dwFlags := LOptions.dwFlags or DTT_CALCRECT;

    with ThemeServices do Result := DrawThemeTextEx(Theme[Details.Element], DC, Details.Part, Details.State,
      PChar(S), Length(S), LFlags, @R, LOptions) = S_OK;
  end
  else
  begin
    // Windows XP doesn't allow changing the text color, use DrawText if needed
    if (stfTextColor in Options.Flags) then
    begin
      LColorRef := SetTextColor(DC, Vcl.Graphics.ColorToRGB(Options.TextColor));
      try
        Winapi.Windows.DrawText(DC, PChar(S), Length(S), R, LFlags);
      finally
        SetTextColor(DC, LColorRef);
      end;
      Result := True;
    end
    else
      // If text color is default for part/state, use DrawThemeText
      with ThemeServices do Result := DrawThemeText(Theme[Details.Element], DC, Details.Part, Details.State,
        PChar(S), Length(S), LFlags, 0, R) = S_OK;
  end;
end;

procedure TSlideSwitch.Paint;
var
  k : integer;
  r : TRect;
  s : string;
  cin,cout,con,coff : TColor;
  Button   : TThemedButton ;
  Details  : TThemedElementDetails;
  Options  : TStyleTextOptionsEx;

  function CenterPos (const AText : string; ARect : TRect) : TPoint;
  var
    w,h : integer;
  begin
    with Canvas do begin
      w:=TextWidth(AText); h:=TextHeight(AText);
      end;
    with Result do begin
      x:=ARect.Left+round((ARect.Width-w)/2);
      y:=ARect.Top+round((ARect.Height-h)/2);
      end;
    end;

begin
  Canvas.Font:=Font;
  with ThemeServices do if ThemesAvailable then begin
    if not Enabled then Button:=tbPushButtonDisabled
    else if FState<>ssOff then Button:=tbPushButtonPressed
    else Button:=tbPushButtonNormal;
    r:=ClientRect;
    Details:=GetElementDetails(Button);
    DrawElement(Canvas.Handle,Details,r);
    if FState=ssOff then begin
      Inc(r.Left,round(2/5*width));
      s:=CaptionOff;
      end
    else begin
      Dec(r.Right,round(2/5*width));
      s:=CaptionOn;
      end;
    with Canvas do begin
      k:=Font.Size div 5;
      Brush.Style:=bsClear;
      with Options do begin
        Flags:=[stfTextColor];
        if Enabled then begin
          if FState=ssOff then begin
            TextColor:=FSwOffColor;
            if k>2 then Flags:=Flags+[stfShadowColor,stfShadowOffset,stfShadowType];
            end
          else begin
            TextColor:=FSwOnColor;
            if k>2 then Flags:=Flags+[stfBorderSize,stfBorderColor];
            end;
          end
        else TextColor:=clWhite;
        BorderColor:=clWhite;
        BorderSize:=k;
        ShadowColor:=clBlack;
        ShadowOffset:=Point(BorderSize div 2,BorderSize);
        ShadowType:=tstSingle;
        GlowSize:=0;
        end;
      DrawTextEx(ThemeServices,Handle,Details,s,r,[tfCenter,tfVerticalCenter,tfSingleLine],Options);
{      if Enabled and (FState<>ssOff) then begin
        OffsetRect(r,0,1);
        Font.Color:=clWhite;
        DrawText(Handle,Details,Caption,r,[tfCenter,tfVerticalCenter,tfSingleLine],Options);
        end;      }
      end;
    if not Enabled then Button:=tbPushButtonDisabled
    else if Focused then Button:=tbPushButtonDefaulted
    else if fMouseHover then Button:=tbPushButtonHot
    else Button:=tbPushButtonNormal;
    r:=fSliderRect;
    Details:=GetElementDetails(Button) ;
    DrawElement(Canvas.Handle,Details,r) ;
    if Focused then begin
      r:=ContentRect(Canvas.Handle,Details,r) ;
      SetTextColor(Canvas.Handle,clWhite);
      DrawFocusRect(Canvas.Handle,r);
     end;
   end
  else begin
    if Enabled then begin
      cout:=FOuterColor;
      cin:=FInnerColor;
      con:=FSwOnColor;
      coff:=FSwOffColor;
      end
    else begin
      cout:=clMedGray;
      cin:=clLtGray;
      con:=clLtGray;
      coff:=clLtGray;
      Canvas.Font.Color:=clWhite;
      end;
    with Canvas do begin
      Brush.Color:=cout;
      with Pen do begin
        Color:=FOuterColor; Width:=0;
        end;
      RoundRect(0,0,Width,Height,3,3);
      Brush.Color:=cin;
      RoundRect(2,2,Width-2,Height-2,3,4);
      Pen.Width:=3;
      if FState=ssOff then begin
        k:=round(2/5*Width);
        Brush.Color:=coff;
        MoveTo(k,2); LineTo(k,Height-2);
        RoundRect(k,3,Width-3,Height-3,3,3);
        r:=Rect(k+1,4,Width-4,Height-4);
        with CenterPos(CaptionOff,r) do TextRect(r,x,y,CaptionOff);
        end
      else begin
        k:=round(3/5*Width);
        Brush.Color:=con;
        MoveTo(k,2); LineTo(k,Height-2);
        RoundRect(3,3,k,Height-3,3,3);
        r:=Rect(4,4,k-1,Height-4);
        with CenterPos(CaptionOn,r) do TextRect(r,x,y,CaptionOn);
        end;
      if Enabled and fMouseHover then begin
        Brush.Color:=FHoverColor;
        Rectangle(fSliderRect);
        end;
      end;
    end;
  end;

//====================================================================
procedure Register;
begin
  RegisterComponents(CompPalPage,[TTumblerSwitch,TMomSwitch,TBiMomSwitch]);
  RegisterComponents(CompPalPage,[TSwitch3D,TMomSwitch3D,TBiMomSwitch3D]);
  RegisterComponents(CompPalPage,[TSlideSwitch]);
end;

end.
